import os
from typing import Any, Dict, Generator, List

import yaml
from clarifai_grpc.grpc.api import resources_pb2, service_pb2
from clarifai_grpc.grpc.api.status import status_code_pb2
from google.protobuf.json_format import MessageToDict
from google.protobuf.wrappers_pb2 import BoolValue

from clarifai.client.app import App
from clarifai.client.base import BaseClient
from clarifai.client.compute_cluster import ComputeCluster
from clarifai.client.lister import Lister
from clarifai.errors import UserError
from clarifai.utils.logging import logger


class User(Lister, BaseClient):
  """User is a class that provides access to Clarifai API endpoints related to user information."""

  def __init__(self,
               user_id: str = None,
               base_url: str = "https://api.clarifai.com",
               pat: str = None,
               token: str = None,
               root_certificates_path: str = None,
               **kwargs):
    """Initializes an User object.

    Args:
        user_id (str): The user ID for the user to interact with.
        base_url (str): Base API url. Default "https://api.clarifai.com"
        pat (str): A personal access token for authentication. Can be set as env var CLARIFAI_PAT
        token (str): A session token for authentication. Accepts either a session token or a pat. Can be set as env var CLARIFAI_SESSION_TOKEN
        root_certificates_path (str): Path to the SSL root certificates file, used to establish secure gRPC connections.
        **kwargs: Additional keyword arguments to be passed to the User.
    """
    self.kwargs = {**kwargs, 'id': user_id}
    self.user_info = resources_pb2.User(**self.kwargs)
    self.logger = logger
    BaseClient.__init__(
        self,
        user_id=self.id,
        app_id="",
        base=base_url,
        pat=pat,
        token=token,
        root_certificates_path=root_certificates_path)
    Lister.__init__(self)

  def list_apps(self, filter_by: Dict[str, Any] = {}, page_no: int = None,
                per_page: int = None) -> Generator[App, None, None]:
    """Lists all the apps for the user.

    Args:
        filter_by (dict): A dictionary of filters to be applied to the list of apps.
        page_no (int): The page number to list.
        per_page (int): The number of items per page.

    Yields:
        App: App objects for the user.

    Example:
        >>> from clarifai.client.user import User
        >>> apps = list(User("user_id").list_apps())

    Note:
        Defaults to 16 per page if page_no is specified and per_page is not specified.
        If both page_no and per_page are None, then lists all the resources.
    """
    request_data = dict(user_app_id=self.user_app_id, **filter_by)
    all_apps_info = self.list_pages_generator(
        self.STUB.ListApps,
        service_pb2.ListAppsRequest,
        request_data,
        per_page=per_page,
        page_no=page_no)
    for app_info in all_apps_info:
      yield App.from_auth_helper(
          self.auth_helper,
          **app_info)  #(base_url=self.base, pat=self.pat, token=self.token, **app_info)

  def list_runners(self, filter_by: Dict[str, Any] = {}, page_no: int = None,
                   per_page: int = None) -> Generator[dict, None, None]:
    """List all runners for the user

    Args:
        filter_by (dict): A dictionary of filters to apply to the list of runners.
        page_no (int): The page number to list.
        per_page (int): The number of items per page.

    Yields:
        Dict: Dictionaries containing information about the runners.

    Example:
        >>> from clarifai.client.user import User
        >>> client = User(user_id="user_id")
        >>> all_runners= list(client.list_runners())

    Note:
        Defaults to 16 per page if page_no is specified and per_page is not specified.
        If both page_no and per_page are None, then lists all the resources.
    """
    request_data = dict(user_app_id=self.user_app_id, **filter_by)
    all_runners_info = self.list_pages_generator(
        self.STUB.ListRunners,
        service_pb2.ListRunnersRequest,
        request_data,
        per_page=per_page,
        page_no=page_no)

    for runner_info in all_runners_info:
      yield dict(auth=self.auth_helper, check_runner_exists=False, **runner_info)

  def list_compute_clusters(self, page_no: int = None,
                            per_page: int = None) -> Generator[dict, None, None]:
    """List all compute clusters for the user

    Args:
        page_no (int): The page number to list.
        per_page (int): The number of items per page.

    Yields:
        Dict: Dictionaries containing information about the compute clusters.

    Example:
        >>> from clarifai.client.user import User
        >>> client = User(user_id="user_id")
        >>> all_compute_clusters= list(client.list_compute_clusters())

    Note:
        Defaults to 16 per page if page_no is specified and per_page is not specified.
        If both page_no and per_page are None, then lists all the resources.
    """
    request_data = dict(user_app_id=self.user_app_id)
    all_compute_clusters_info = self.list_pages_generator(
        self.STUB.ListComputeClusters,
        service_pb2.ListComputeClustersRequest,
        request_data,
        per_page=per_page,
        page_no=page_no)

    for compute_cluster_info in all_compute_clusters_info:
      yield ComputeCluster.from_auth_helper(self.auth_helper, **compute_cluster_info)

  def create_app(self, app_id: str, base_workflow: str = 'Empty', **kwargs) -> App:
    """Creates an app for the user.

    Args:
        app_id (str): The app ID for the app to create.
        base_workflow (str): The base workflow to use for the app.(Examples: 'Universal', 'Language-Understanding', 'General')
        **kwargs: Additional keyword arguments to be passed to the App.

    Returns:
        App: An App object for the specified app ID.

    Example:
        >>> from clarifai.client.user import User
        >>> client = User(user_id="user_id")
        >>> app = client.create_app(app_id="app_id",base_workflow="Universal")
    """
    workflow = resources_pb2.Workflow(id=base_workflow, app_id="main", user_id="clarifai")
    request = service_pb2.PostAppsRequest(
        user_app_id=self.user_app_id,
        apps=[resources_pb2.App(id=app_id, default_workflow=workflow, **kwargs)])
    response = self._grpc_request(self.STUB.PostApps, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    self.logger.info("\nApp created\n%s", response.status)
    return App.from_auth_helper(auth=self.auth_helper, app_id=app_id)

  def create_runner(self, runner_id: str, labels: List[str], description: str) -> dict:
    """Create a runner

    Args:
      runner_id (str): The Id of runner to create
      labels (List[str]): Labels to match runner
      description (str): Description of Runner

    Returns:
      Dict: A dictionary containing information about the specified Runner ID.

    Example:
        >>> from clarifai.client.user import User
        >>> client = User(user_id="user_id")
        >>> runner_info = client.create_runner(runner_id="runner_id", labels=["label to link runner"], description="laptop runner")
    """

    if not isinstance(labels, List):
      raise UserError("Labels must be a List of strings")

    request = service_pb2.PostRunnersRequest(
        user_app_id=self.user_app_id,
        runners=[resources_pb2.Runner(id=runner_id, labels=labels, description=description)])
    response = self._grpc_request(self.STUB.PostRunners, request)

    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    self.logger.info("\nRunner created\n%s", response.status)

    return dict(
        auth=self.auth_helper,
        runner_id=runner_id,
        user_id=self.id,
        labels=labels,
        description=description,
        check_runner_exists=False)

  def _process_compute_cluster_config(self, config_filepath: str) -> Dict[str, Any]:
    with open(config_filepath, "r") as file:
      compute_cluster_config = yaml.safe_load(file)

    assert "compute_cluster" in compute_cluster_config, "compute cluster info not found in the config file"
    compute_cluster = compute_cluster_config['compute_cluster']
    assert "region" in compute_cluster, "region not found in the config file"
    assert "managed_by" in compute_cluster, "managed_by not found in the config file"
    assert "cluster_type" in compute_cluster, "cluster_type not found in the config file"
    compute_cluster['cloud_provider'] = resources_pb2.CloudProvider(
        **compute_cluster['cloud_provider'])
    compute_cluster['key'] = resources_pb2.Key(id=self.pat)
    if "visibility" in compute_cluster:
      compute_cluster["visibility"] = resources_pb2.Visibility(**compute_cluster["visibility"])
    return compute_cluster

  def create_compute_cluster(self, config_filepath: str,
                             compute_cluster_id: str = None) -> ComputeCluster:
    """Creates a compute cluster for the user.

    Args:
        config_filepath (str): The path to the compute cluster config file.
        compute_cluster_id (str): New compute cluster ID for the compute cluster to create.

    Returns:
        ComputeCluster: A Compute Cluster object for the specified compute cluster ID.

    Example:
        >>> from clarifai.client.user import User
        >>> client = User(user_id="user_id")
        >>> compute_cluster = client.create_compute_cluster(config_filepath="config.yml")
    """
    if not os.path.exists(config_filepath):
      raise UserError(f"Compute Cluster config file not found at {config_filepath}")

    compute_cluster_config = self._process_compute_cluster_config(config_filepath)

    if 'id' in compute_cluster_config:
      if compute_cluster_id is None:
        compute_cluster_id = compute_cluster_config['id']
      compute_cluster_config.pop('id')

    request = service_pb2.PostComputeClustersRequest(
        user_app_id=self.user_app_id,
        compute_clusters=[
            resources_pb2.ComputeCluster(id=compute_cluster_id, **compute_cluster_config)
        ])
    response = self._grpc_request(self.STUB.PostComputeClusters, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    self.logger.info("\nCompute Cluster created\n%s", response.status)
    return ComputeCluster.from_auth_helper(
        auth=self.auth_helper, compute_cluster_id=compute_cluster_id)

  def app(self, app_id: str, **kwargs) -> App:
    """Returns an App object for the specified app ID.

    Args:
        app_id (str): The app ID for the app to interact with.
        **kwargs: Additional keyword arguments to be passed to the App.

    Returns:
        App: An App object for the specified app ID.

    Example:
        >>> from clarifai.client.user import User
        >>> app = User("user_id").app("app_id")
    """
    request = service_pb2.GetAppRequest(
        user_app_id=resources_pb2.UserAppIDSet(user_id=self.id, app_id=app_id))
    response = self._grpc_request(self.STUB.GetApp, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)

    kwargs['user_id'] = self.id
    return App.from_auth_helper(auth=self.auth_helper, app_id=app_id, **kwargs)

  def runner(self, runner_id: str) -> dict:
    """Returns a Runner object if exists.

    Args:
        runner_id (str): The runner ID to interact with

    Returns:
        Dict: A dictionary containing information about the existing runner ID.

    Example:
        >>> from clarifai.client.user import User
        >>> client = User(user_id="user_id")
        >>> runner_info = client.runner(runner_id="runner_id")
    """
    request = service_pb2.GetRunnerRequest(user_app_id=self.user_app_id, runner_id=runner_id)
    response = self._grpc_request(self.STUB.GetRunner, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(
          f"""Error getting runner, are you use this is a valid runner id {runner_id} at the user_id/app_id
          {self.user_app_id.user_id}/{self.user_app_id.app_id}.
          Error: {response.status.description}""")

    dict_response = MessageToDict(response, preserving_proto_field_name=True)
    kwargs = self.process_response_keys(dict_response[list(dict_response.keys())[1]],
                                        list(dict_response.keys())[1])

    return dict(self.auth_helper, check_runner_exists=False, **kwargs)

  def compute_cluster(self, compute_cluster_id: str) -> ComputeCluster:
    """Returns an Compute Cluster object for the specified compute cluster ID.

    Args:
        compute_cluster_id (str): The compute cluster ID for the compute cluster to interact with.

    Returns:
        ComputeCluster: A Compute Cluster object for the specified compute cluster ID.

    Example:
        >>> from clarifai.client.user import User
        >>> compute_cluster = User("user_id").compute_cluster("compute_cluster_id")
    """
    request = service_pb2.GetComputeClusterRequest(
        user_app_id=self.user_app_id, compute_cluster_id=compute_cluster_id)
    response = self._grpc_request(self.STUB.GetComputeCluster, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)

    dict_response = MessageToDict(response, preserving_proto_field_name=True)
    kwargs = self.process_response_keys(dict_response[list(dict_response.keys())[1]],
                                        list(dict_response.keys())[1])

    return ComputeCluster.from_auth_helper(auth=self.auth_helper, **kwargs)

  def patch_app(self, app_id: str, action: str = 'overwrite', **kwargs) -> App:
    """Patch an app for the user.

    Args:
        app_id (str): The app ID for the app to patch.
        action (str): The action to perform on the app (overwrite/remove).
        **kwargs: Additional keyword arguments to be passed to patch the App.

    Returns:
        App: Patched App object for the specified app ID.
    """
    if "base_workflow" in kwargs:
      kwargs["default_workflow"] = resources_pb2.Workflow(
          id=kwargs.pop("base_workflow"), app_id="main", user_id="clarifai")
    if "visibility" in kwargs:
      kwargs["visibility"] = resources_pb2.Visibility(gettable=kwargs["visibility"])
    if "image_url" in kwargs:
      kwargs["image"] = resources_pb2.Image(url=kwargs.pop("image_url"))
    if "is_template" in kwargs:
      kwargs["is_template"] = BoolValue(value=kwargs["is_template"])
    request = service_pb2.PatchAppRequest(
        user_app_id=resources_pb2.UserAppIDSet(user_id=self.id, app_id=app_id),
        app=resources_pb2.App(id=app_id, **kwargs),
        action=action,
        reindex=False)
    response = self._grpc_request(self.STUB.PatchApp, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    self.logger.info("\nApp patched\n%s", response.status)

    return App.from_auth_helper(auth=self.auth_helper, app_id=app_id)

  def delete_app(self, app_id: str) -> None:
    """Deletes an app for the user.

    Args:
        app_id (str): The app ID for the app to delete.

    Example:
        >>> from clarifai.client.user import User
        >>> user = User("user_id").delete_app("app_id")
    """
    request = service_pb2.DeleteAppRequest(
        user_app_id=resources_pb2.UserAppIDSet(user_id=self.id, app_id=app_id))
    response = self._grpc_request(self.STUB.DeleteApp, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    self.logger.info("\nApp Deleted\n%s", response.status)

  def delete_runner(self, runner_id: str) -> None:
    """Deletes all spectified runner ids

    Args:
        runner_ids (str): List of runners to delete

    Example:
        >>> from clarifai.client.user import User
        >>> client = User(user_id="user_id")
        >>> client.delete_runner(runner_id="runner_id")
    """
    request = service_pb2.DeleteRunnersRequest(user_app_id=self.user_app_id, ids=[runner_id])
    response = self._grpc_request(self.STUB.DeleteRunners, request)

    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    self.logger.info("\nRunner Deleted\n%s", response.status)

  def delete_compute_clusters(self, compute_cluster_ids: List[str]) -> None:
    """Deletes a list of compute clusters for the user.

    Args:
        compute_cluster_ids (List[str]): The compute cluster IDs of the user to delete.

    Example:
        >>> from clarifai.client.user import User
        >>> user = User("user_id").delete_compute_clusters(compute_cluster_ids=["compute_cluster_id1", "compute_cluster_id2"])
    """
    assert isinstance(compute_cluster_ids, list), "compute_cluster_ids param should be a list"

    request = service_pb2.DeleteComputeClustersRequest(
        user_app_id=self.user_app_id, ids=compute_cluster_ids)
    response = self._grpc_request(self.STUB.DeleteComputeClusters, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    self.logger.info("\nCompute Cluster Deleted\n%s", response.status)

  def __getattr__(self, name):
    return getattr(self.user_info, name)

  def __str__(self):
    init_params = [param for param in self.kwargs.keys()]
    attribute_strings = [
        f"{param}={getattr(self.user_info, param)}" for param in init_params
        if hasattr(self.user_info, param)
    ]
    return f"Clarifai User Details: \n{', '.join(attribute_strings)}\n"



import click
from clarifai.cli.base import cli
from clarifai.client.user import User
from clarifai.utils.cli import display_co_resources


@cli.group(['computecluster', 'cc'])
def computecluster():
  """Manage Compute Clusters: create, delete, list"""
  pass


@computecluster.command()
@click.option(
    '--config',
    type=click.Path(exists=True),
    required=True,
    help='Path to the compute cluster config file.')
@click.option(
    '-cc_id',
    '--compute_cluster_id',
    required=False,
    help='New Compute Cluster ID for the compute cluster to create.')
@click.pass_context
def create(ctx, config, compute_cluster_id):
  """Create a new Compute Cluster with the given config file."""
  user = User(user_id=ctx.obj['user_id'], pat=ctx.obj['pat'], base_url=ctx.obj['base_url'])
  if compute_cluster_id:
    user.create_compute_cluster(config, compute_cluster_id=compute_cluster_id)
  else:
    user.create_compute_cluster(config)


@computecluster.command()
@click.option('--page_no', required=False, help='Page number to list.', default=1)
@click.option('--per_page', required=False, help='Number of items per page.', default=16)
@click.pass_context
def list(ctx, page_no, per_page):
  """List all compute clusters for the user."""
  user = User(user_id=ctx.obj['user_id'], pat=ctx.obj['pat'], base_url=ctx.obj['base_url'])
  response = user.list_compute_clusters(page_no, per_page)
  display_co_resources(response, "Compute Cluster")


@computecluster.command()
@click.option('-cc_id', '--compute_cluster_id', help='Compute Cluster ID of the user to delete.')
@click.pass_context
def delete(ctx, compute_cluster_id):
  """Deletes a compute cluster for the user."""
  user = User(user_id=ctx.obj['user_id'], pat=ctx.obj['pat'], base_url=ctx.obj['base_url'])
  user.delete_compute_clusters([compute_cluster_id])



from .models.base_typed_model import AnyAnyModel, TextInputModel, VisualInputModel
from .models.model_builder import ModelBuilder
from .models.model_runner import ModelRunner
from .utils.data_handler import InputDataHandler, OutputDataHandler

__all__ = [
    "ModelRunner",
    "ModelBuilder",
    "InputDataHandler",
    "OutputDataHandler",
    "AnyAnyModel",
    "TextInputModel",
    "VisualInputModel",
]



from schema import And, Optional, Regex, Schema, SchemaError, Use

# Non-empty, up to 32-character ASCII strings with internal dashes and underscores.
_id_validator = And(str, lambda s: 0 < len(s) <= 48, Regex(r'^[0-9A-Za-z]+([-_][0-9A-Za-z]+)*$'))

# 32-character hex string, converted to lower-case.
_hex_id_validator = And(str, Use(str.lower), Regex(r'^[0-9a-f]{32}'))


def _model_does_not_have_model_version_id_and_other_fields(m):
  """ Validate that model does not have model_version_id and other model fields."""
  if ('model_version_id' in m) and _model_has_other_fields(m):
    raise SchemaError(f"model should not set model_version_id and other model fields: {m};"
                      f" please remove model_version_id or other model fields.")
  return True


def _model_has_other_fields(m):
  return any(k not in ['model_id', 'model_version_id', 'user_id', 'app_id'] for k in m.keys())


def _workflow_nodes_have_valid_dependencies(nodes):
  """Validate that all inputs to a node are declared before it."""
  node_ids = set()
  for node in nodes:
    for node_input in node.get("node_inputs", []):
      if node_input["node_id"] not in node_ids:
        raise SchemaError(f"missing input '{node_input['node_id']}' for node '{node['id']}'")
    node_ids.add(node["id"])

  return True


_data_schema = Schema({
    "workflow": {
        "id":
            _id_validator,
        "nodes":
            And(
                len,
                [{
                    "id":
                        And(str, len),  # Node IDs are not validated as IDs by the API.
                    "model":
                        And({
                            "model_id": _id_validator,
                            Optional("app_id"): _id_validator,
                            Optional("user_id"): _id_validator,
                            Optional("model_version_id"): _hex_id_validator,
                            Optional("model_type_id"): _id_validator,
                            Optional("description"): str,
                            Optional("output_info"): {
                                Optional("params"): dict,
                            },
                        }, _model_does_not_have_model_version_id_and_other_fields),
                    Optional("node_inputs"):
                        And(len, [{
                            "node_id": And(str, len),
                        }]),
                }],
                _workflow_nodes_have_valid_dependencies),
    },
})


def validate(data):
  return _data_schema.validate(data)



import os
import typing
import uuid

import pytest
from google.protobuf import struct_pb2

from clarifai.client.search import Search
from clarifai.client.user import User
from clarifai.errors import UserError

CREATE_APP_USER_ID = os.environ["CLARIFAI_USER_ID"]
uniq = uuid.uuid4().hex[:10]
CREATE_APP_ID = f"ci_search_app_{uniq}"
CREATE_DATASET_ID = "ci_search_dataset"
DOG_IMG_URL = "https://samples.clarifai.com/dog.tiff"
DATASET_IMAGES_DIR = os.path.dirname(__file__) + "/assets/voc/images"

CLARIFAI_API_BASE = os.environ.get("CLARIFAI_API_BASE", "https://api.clarifai.com")


def get_filters_for_test() -> [(typing.List[typing.Dict], int)]:
  return [
      ([{
          "geo_point": {
              "longitude": -29.0,
              "latitude": 40.0,
              "geo_limit": 100
          }
      }], 1),
      ([{
          "concepts": [{
              "name": "dog",
              "value": 1
          }]
      }], 1),
      (
          [{  # OR
              "concepts": [{
                  "name": "deer",
                  "value": 1
              }, {
                  "name": "dog",
                  "value": 1
              }]
          }],
          1),
      (
          [
              {  # AND
                  "concepts": [{
                      "name": "dog",
                      "value": 1
                  }]
              },
              {
                  "concepts": [{
                      "name": "deer",
                      "value": 1
                  }]
              }
          ],
          0),
    (
        [

                {
                "metadata": {"Breed": "Saint Bernard"}
                }
        ],
        1),

      # Input Search
      (
            [
                { # AND
                  "input_types": ["image"],
                },
                {
                  "input_status_code": 30000 # Download Success
                }
            ],
            1),
      (
            [
                {
                  "input_types": ["text", "audio", "video"],
                }
            ],
            0),
      (
            [
                { # OR
                  "input_types": ["text", "audio", "video"],
                  "input_status_code": 30000 # Download Success
                },
            ],
            1),
      (
            [
                {
                  "input_dataset_ids": ["random_dataset"]
                },
            ],
            0),
  ]


@pytest.mark.requires_secrets
class TestAnnotationSearch:

  @classmethod
  def setup_class(cls):
    cls.client = User(user_id=CREATE_APP_USER_ID, base_url=CLARIFAI_API_BASE)
    cls.search = Search(
        user_id=CREATE_APP_USER_ID,
        app_id=CREATE_APP_ID,
        base_url=CLARIFAI_API_BASE,
        top_k=1,
        metric="euclidean")
    cls.search_with_pagination = Search(
        user_id=CREATE_APP_USER_ID,
        app_id=CREATE_APP_ID,
        base_url=CLARIFAI_API_BASE,
        metric="euclidean",
        pagination=True)
    cls.search_deduplicate = Search(
        user_id=CREATE_APP_USER_ID,
        app_id=CREATE_APP_ID,
        base_url=CLARIFAI_API_BASE,
        top_k=2,
        metric="euclidean")
    cls.upload_data()

  @classmethod
  def upload_data(self):
    app_obj = self.client.create_app(CREATE_APP_ID, base_workflow="General")
    dataset_obj = app_obj.create_dataset(CREATE_DATASET_ID)
    inp_obj = app_obj.inputs()
    metadata = struct_pb2.Struct()
    metadata.update({"Breed": "Saint Bernard"})
    input_proto = inp_obj.get_input_from_url(
        dataset_id=CREATE_DATASET_ID,
        input_id="dog-tiff",
        image_url=DOG_IMG_URL,
        labels=["dog"],
        geo_info=[-30.0, 40.0],  # longitude, latitude
        metadata=metadata)
    inp_obj.upload_inputs([input_proto])
    dataset_obj.upload_from_folder(DATASET_IMAGES_DIR, input_type="image", labels=False)

  @pytest.mark.parametrize("filter_dict_list,expected_hits", get_filters_for_test())
  def test_filter_search(self, filter_dict_list: typing.List[typing.Dict], expected_hits: int):
    query = self.search.query(filters=filter_dict_list)
    assert len(list(query)) == expected_hits

  def test_rank_search(self):
    query = self.search.query(ranks=[{"image_url": "https://samples.clarifai.com/dog.tiff"}])
    for q in query:
      assert len(q.hits) == 1
      assert q.hits[0].input.id == "dog-tiff"

  def test_rank_filter_search(self):
    query = self.search.query(
        ranks=[{
            "image_url": "https://samples.clarifai.com/dog.tiff"
        }],
        filters=[{
            "input_types": ["image"]
        }])
    for q in query:
      assert len(q.hits) == 1
      assert q.hits[0].input.id == "dog-tiff"

  def test_per_page(self):
    query = self.search_with_pagination.query(
        filters=[{
            "input_types": ["image"]
        }], per_page=3, page_no=1)
    for q in query:
      assert len(q.hits) == 3

  def test_pagination(self):
    query = self.search_with_pagination.query(filters=[{"input_types": ["image"]}])
    for q in query:
      assert len(q.hits) == 11

  def test_schema_error(self):
    with pytest.raises(UserError):
      _ = self.search.query(filters=[{
          "geo_point": {
              "longitude": -29.0,
              "latitude": 40.0,
              "geo_limit": 10,
              "extra": 1
          }
      }])

    # Incorrect Concept Keys
    with pytest.raises(UserError):
      _ = self.search.query(filters=[{
          "concepts": [{
              "value": 1,
              "concept_id": "deer"
          }, {
              "name": "dog",
              "value": 1
          }]
      }])

    # Incorrect Concept Values
    with pytest.raises(UserError):
      _ = self.search.query(filters=[{
          "concepts": [{
              "name": "deer",
              "value": 2
          }, {
              "name": "dog",
              "value": 1
          }]
      }])

    # Incorrect input type search
    with pytest.raises(UserError):
      _ = self.search.query(filters=[{"input_types": ["imaage"]}])

    # Incorrect input search filter key
    with pytest.raises(UserError):
      _ = self.search.query(filters=[{"input_id": "test"}])

  def test_rank_search_deduplicate(self):
    query = self.search_deduplicate.query(ranks=[{
        "image_url": "https://samples.clarifai.com/dog.tiff"
    }])
    for q in query:
      assert len(q.hits) == 2
      assert q.hits[0].input.id != q.hits[1].input.id

  def teardown_class(cls):
    cls.client.delete_app(app_id=CREATE_APP_ID)



from io import BytesIO

from PIL import Image


def image_to_bytes(img: Image.Image, format="JPEG") -> bytes:
  buffered = BytesIO()
  img.save(buffered, format=format)
  img_str = buffered.getvalue()
  return img_str


def bytes_to_image(bytes_img) -> Image.Image:
  img = Image.open(BytesIO(bytes_img))
  return img
